import textwrap as tw

import pytest

from zgulde.flashcards.__main__ import FlashCard, make_cards


def test_card_parsing():
    content = tw.dedent(
        """
    here is the front
    ---
    and this is the back
    ===
    this is another card
    ---
    this is the second back
    """
    ).strip()

    cards = make_cards(content)

    assert len(cards) == 2

    first, second = cards

    assert first.front == "here is the front"
    assert first.back == "and this is the back"

    assert second.front == "this is another card"
    assert second.back == "this is the second back"


def test_whitespace_doesnt_matter_in_card_parsing():
    content = tw.dedent(
        """

    here is the front

    ---

    and this is the back

    ===

    this is another card


    ---

    this is the second back
    """
    ).strip()

    cards = make_cards(content)

    assert len(cards) == 2

    first, second = cards

    assert first.front == "here is the front"
    assert first.back == "and this is the back"

    assert second.front == "this is another card"
    assert second.back == "this is the second back"



"""
python -m zgulde.em --help

TODO:

- preserve indentation after doc marker comments (e.g. multiline list items)
"""
import fileinput
import itertools as it
import sys
import textwrap

FILE_EXTENSIONS = dict(
    sh=dict(comment="#", language="bash"),
    py=dict(comment="#", language="python"),
    js=dict(comment="//", language="javascript"),
    sql=dict(comment="--", language="sql"),
    php=dict(comment="//", language="php"),
)


def prev_and_next(xs):
    prevs, items, nexts = it.tee(xs, 3)
    return zip(
        it.chain([None], prevs), items, it.chain(it.islice(nexts, 1, None), [None])
    )


def make_categorizor(doc_marker):
    def is_code(line):
        return line is not None and not line.startswith(doc_marker) and line != ""

    def is_docs(line):
        return line is not None and line.startswith(doc_marker)

    def categorize(x):
        prev, line, next = x
        if is_docs(line):
            return "docs"
        elif is_code(line):
            return "code"
        elif line == "" and is_code(prev) and is_code(next):
            return "code"
        elif line == "" and is_docs(prev) and is_docs(next):
            return "docs"
        else:
            return "ignore"

    return categorize


def main(fp, language, doc_marker):
    lines = map(lambda line: line.rstrip(), fileinput.input(fp))
    keyfn = make_categorizor(doc_marker)
    chars_to_skip = len(doc_marker)

    for section_type, contents in it.groupby(prev_and_next(lines), keyfn):
        if section_type == "ignore":
            pass
        elif section_type == "code":
            print(f"```{language}")
            for _, line, _ in contents:
                print(line)
            print("```")
            print()
        elif section_type == "docs":
            for _, line, _ in contents:
                print(line[chars_to_skip:].lstrip())
            print()
        else:
            raise ValueError(f"Unknown section type: {section_type}")


def get_defaults(args):
    fp = args.file if args.file is not None else "-"
    extension = fp.split(".")[-1]
    language = args.language
    doc_marker = args.doc_marker

    if language is None and extension in FILE_EXTENSIONS.keys():
        language = FILE_EXTENSIONS[extension]["language"]
    elif language is None:
        language = ""

    if doc_marker is None and extension in FILE_EXTENSIONS.keys():
        doc_marker = FILE_EXTENSIONS[extension]["comment"] + ":"
    elif doc_marker is None:
        doc_marker = "#:"

    return fp, language, doc_marker


def show_defaults():
    print("ext marker language")
    for extension, d in FILE_EXTENSIONS.items():
        print("{:>3} {:^6} {}".format(extension, d["comment"] + ":", d["language"]))


if __name__ == "__main__":
    import argparse
    import textwrap

    description = """
    Transform source code -> markdown.

    Specially marked lines (those that start with the doc_marker) are treated as
    markdown, and everything else is turned into a fenced code block.
    """

    parser = argparse.ArgumentParser()
    parser.prog = "python -m zgulde.em"
    parser.description = textwrap.dedent(description)
    parser.add_argument("file", metavar="FILE", help="filepath", nargs="?")
    parser.add_argument("-m", "--doc-marker", help="indicator for a documentation line")
    parser.add_argument("-l", "--language", help="language for fenced code blocks")
    parser.add_argument(
        "--show-defaults",
        action="store_true",
        help="show defaults for each file extension and exit",
    )
    args = parser.parse_args()

    if args.show_defaults:
        show_defaults()
        sys.exit(0)

    fp, language, doc_marker = get_defaults(args)
    main(fp, language, doc_marker)



import git
from tqdm import tqdm


def to_dict(commit):
    subject, *body = commit.message.split("\n")
    body = "\n".join(body).strip()
    return {
        "author_email": commit.author.email,
        "author_name": commit.author.name,
        "authored_at": str(commit.authored_datetime),
        "body": body,
        "committed_at": str(commit.committed_datetime),
        "committer_email": commit.committer.email,
        "committer_name": commit.committer.name,
        "deletions": commit.stats.total["deletions"],
        "files_changed": commit.stats.total["files"],
        "insertions": commit.stats.total["insertions"],
        "lines_changed": commit.stats.total["lines"],
        "message": commit.message,
        "n_parents": len(commit.parents),
        "parents": [parent.hexsha for parent in commit.parents],
        "sha": commit.hexsha,
        "size": commit.size,
        "subject": subject,
        "type": commit.type,
    }


def get_commits(repo_path):
    repo = git.Repo(repo_path)
    commits = [to_dict(commit) for commit in tqdm(list(repo.iter_commits()))]
    return commits


def get_commits_df(repo_path):
    import pandas as pd

    df = pd.DataFrame(get_commits(repo_path))
    df.committed_at = pd.to_datetime(df.committed_at, utc=True)
    df.authored_at = pd.to_datetime(df.authored_at, utc=True)
    return df



"""
A cool little module for visually showing regular expression match results.
Intented for interactive regexp iteration.

There are 3 main sets of functions for viewing matches in different
environments: jupyter notebooks, command line interfaces, and plaintext.

In addition to the 3 different environments, each function comes in 2 flavors
for highlighting all of the matches, or just the first match in the string.

- For notebooks: `hl_matches_nb` and `hl_all_matches_nb`
- For the command line: `hl_matches_cli` and `hl_all_matches_cli`
- For plaintext results: `hl_all_matches_plaintext` and `hl_all_matches_plaintext`

For example:

>>> from zgulde.hl_matches import hl_all_matches_plaintext as hl
>>> # square brackets, [], mark the overall match start and end
>>> # parenthesis, (), mark capture group starts and ends
>>> hl(r'.+', 'abcdef')
'[abcdef]'
>>> hl(r'\w(\w{2})', 'abcdef')
'[a(bc)][d(ef)]'

Note that currently, no special highlighting is done for nested groups.
"""
import re
from typing import Iterator, List, Match, Tuple

from IPython.display import HTML as nb_HTML
from IPython.display import display
from prompt_toolkit import HTML as cli_HTML
from prompt_toolkit import print_formatted_text

MatchIndexes = List[int]


def _starts_ends_overall(m: Match) -> Tuple[MatchIndexes, MatchIndexes, MatchIndexes]:
    """
    Extracts indices from a match object.

    Returns

    (groupstarts, groupends, [overall_start, overall_end])

    >>> m = re.match(r'.(.)', 'abc')
    >>> _starts_ends_overall(m)
    ([1], [2], [0, 2])
    >>> m = re.match(r'.', 'abc')
    >>> _starts_ends_overall(m)
    ([], [], [0, 1])
    """
    overall_start, overall_end = m.span()
    n_matches = len(m.groups())

    spans = [m.span(n) for n in range(1, n_matches + 1)]
    starts = [span[0] for span in spans]
    ends = [span[1] for span in spans]
    return starts, ends, [overall_start, overall_end]


def _starts_ends_overalls(
    matches: Iterator[Match]
) -> Tuple[MatchIndexes, MatchIndexes, MatchIndexes, MatchIndexes]:
    """
    >>> matches = re.finditer(r'.', 'abc')
    >>> _starts_ends_overalls(matches)
    ([], [], [0, 1, 2], [1, 2, 3])
    >>> matches = re.finditer(r'.(.)', 'abc')
    >>> _starts_ends_overalls(matches)
    ([1], [2], [0], [2])
    >>> matches = re.finditer(r'.(.)', 'abcdef')
    >>> _starts_ends_overalls(matches)
    ([1, 3, 5], [2, 4, 6], [0, 2, 4], [2, 4, 6])
    """
    groupstarts: List[int] = []
    groupends: List[int] = []
    overallstarts: List[int] = []
    overallends: List[int] = []
    for starts, ends, (start, end) in map(_starts_ends_overall, matches):
        groupstarts += starts
        groupends += ends
        overallstarts += [start]
        overallends += [end]
    return groupstarts, groupends, overallstarts, overallends


def _hl_all_matches(
    regexp: str,
    subject: str,
    start: str,
    end: str,
    groupstart: str,
    groupend: str,
    initial_output="",
    closing_output="",
) -> str:
    """
    Does not handle nested groups.
    """

    matches = re.finditer(regexp, subject)

    groupstarts, groupends, overallstarts, overallends = _starts_ends_overalls(matches)

    output = initial_output

    for i, c in enumerate(subject):
        if i in groupends:
            output += groupend
        if i in overallends:
            output += end
        if i in overallstarts:
            output += start
        if i in groupstarts:
            output += groupstart

        output += c

    if len(subject) in groupends:
        output += groupend
    if len(subject) in overallends:
        output += end

    output += closing_output
    return output


def _hl_matches(
    regexp: str,
    subject: str,
    start: str,
    end: str,
    groupstart: str,
    groupend: str,
    initial_output="",
    closing_output="",
) -> str:
    """
    Does not handle nested groups.
    """

    m = re.search(regexp, subject)
    if m is None:
        return subject

    overall_start, overall_end = m.span()
    n_matches = len(m.groups())

    spans = [m.span(n) for n in range(1, n_matches + 1)]
    starts = [span[0] for span in spans]
    ends = [span[1] for span in spans]

    output = initial_output

    for i, c in enumerate(subject):
        if i == overall_start:
            output += start

        if i in starts:
            output += groupstart

        if i in ends:
            output += groupend

        if i == overall_end:
            output += end

        output += c

    if len(subject) in ends:
        output += groupend
    if len(subject) == overall_end:
        output += end

    output += closing_output
    return output


def hl_matches_cli(regexp, subject):
    """
    Highlight a regular expressions first match in a string. The entire match
    will be underlined, the capture groups will be red.

    This function only works in terminal environments (read: this function won't
    work in a jupyter notebook).

    Uses python's `re.search` under the hood.
    """
    output = _hl_matches(
        regexp,
        subject,
        start="<u>",
        end="</u>",
        groupstart="<firebrick>",
        groupend="</firebrick>",
    )
    print_formatted_text(cli_HTML(output))


def hl_matches_nb(regexp, subject):
    """
    Highlight a regular expressions first match in a string. The entire match will
    be underlined, and the capture groups will be red.

    This function will only work in jupyter notebooks.

    Uses python's `re.search` under the hood.
    """
    output = _hl_matches(
        regexp,
        subject,
        start='<span style="text-decoration: underline">',
        end="</span>",
        groupstart='<span style="color: firebrick">',
        groupend="</span>",
        initial_output='<div style="font-family: monospace; letter-spacing: 3px; font-size: 24px; line-height: 36px;">',
        closing_output="</div>",
    )
    return nb_HTML(output)


def hl_matches_plaintext(regexp, subject):
    """
    Highlightes a regular expression's first match with plaintext markers.

    >>> from zgulde.hl_matches import hl_matches_plaintext as hl
    >>> hl(r'.+', 'abc')
    '[abc]'
    >>> hl(r'.', 'abc')
    '[a]bc'
    >>> hl(r'.(.).', 'abc')
    '[a(b)c]'
    >>> hl(r'^(\d+).*?(\d+)$', '123 broadway st san antonio tx 78205')
    '[(123) broadway st san antonio tx (78205)]'
    """
    return _hl_matches(
        regexp,
        subject,
        start="[",
        end="]",
        groupstart="(",
        groupend=")",
        initial_output="",
        closing_output="",
    )


def hl_all_matches_plaintext(regexp, subject):
    """
    Highlightes all of a regular expression's matches in `subject` with
    plaintext markers.

    >>> from zgulde.hl_matches import hl_all_matches_plaintext as hl
    >>> hl(r'.+', 'abc')
    '[abc]'
    >>> hl(r'.', 'abc')
    '[a][b][c]'
    >>> hl(r'..', 'abc')
    '[ab]c'
    >>> hl(r'.(.).', 'abc')
    '[a(b)c]'
    >>> hl(r'^(\d+).*?(\d+)$', '123 broadway st san antonio tx 78205')
    '[(123) broadway st san antonio tx (78205)]'
    >>> hl(r'^.', 'abc')
    '[a]bc'
    """
    return _hl_all_matches(
        regexp,
        subject,
        start="[",
        end="]",
        groupstart="(",
        groupend=")",
        initial_output="",
        closing_output="",
    )


def hl_all_matches_nb(regexp, subject):
    """
    Highlight a regular expressions matches in a string. Entire matches will be
    underlined, and capture groups will be red.

    This function will only work in jupyter notebooks.

    Uses python's `re.finditer` under the hood.
    """
    output = _hl_all_matches(
        regexp,
        subject,
        start='<span style="text-decoration: underline">',
        end="</span>",
        groupstart='<span style="color: firebrick">',
        groupend="</span>",
        initial_output='<div style="font-family: monospace; letter-spacing: 3px; font-size: 24px; line-height: 36px;">',
        closing_output="</div>",
    )
    return nb_HTML(output)


def hl_all_matches_cli(regexp, subject):
    """
    Highlight a regular expressions matches in a string. Entire matches will be
    underlined, and capture groups will be red.

    This function will only work in command line interfaces (read: not in
    jupyter notebooks).

    Uses python's `re.finditer` under the hood.
    """
    output = _hl_all_matches(
        regexp,
        subject,
        start="<u>",
        end="</u>",
        groupstart="<firebrick>",
        groupend="</firebrick>",
    )
    print_formatted_text(cli_HTML(output))



import re
from typing import Iterable, Tuple, Union

import nltk
import pandas as pd
import sklearn
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer
from toolz import comp, partial


def tfidfs(
    documents: Iterable[str], return_vectorizer=False
) -> Union[
    pd.DataFrame, Tuple[pd.DataFrame, sklearn.feature_extraction.text.TfidfVectorizer]
]:
    """
    >>> docs = [
    ...     "doc number one",
    ...     "this is the second one",
    ...     "this is the third doc"
    ... ]
    >>> tfidfs(docs)
            doc        is    number       one    second       the     third      this
    0  0.517856  0.000000  0.680919  0.517856  0.000000  0.000000  0.000000  0.000000
    1  0.000000  0.417796  0.000000  0.417796  0.549351  0.417796  0.000000  0.417796
    2  0.417796  0.417796  0.000000  0.000000  0.000000  0.417796  0.549351  0.417796
    """
    vectorizer = TfidfVectorizer()
    m = vectorizer.fit_transform(documents).todense()
    df = pd.DataFrame(m, columns=vectorizer.get_feature_names())
    if return_vectorizer:
        return df, vectorizer
    else:
        return df


def bag_of_words(
    documents: Iterable[str], return_cv=False
) -> Union[
    pd.DataFrame, Tuple[pd.DataFrame, sklearn.feature_extraction.text.CountVectorizer]
]:
    """
    >>> docs = ["first article", "second article", "mary had a little lamb, little lamb"]
    >>> bag_of_words(docs)
       article  first  had  lamb  little  mary  second
    0        1      1    0     0       0     0       0
    1        1      0    0     0       0     0       1
    2        0      0    1     2       2     1       0
    """
    cv = CountVectorizer()
    m = cv.fit_transform(documents).todense()
    bow = pd.DataFrame(m, columns=cv.get_feature_names())
    if return_cv:
        return bow, cv
    else:
        return bow


def top_n_ngrams(s: pd.Series, top_n=3, ngrams=3):
    """
    Extract the top_n number of ngrams (n=ngrams) from s. s is assumed to be a
    pandas Series full of text data.

    >>> s = pd.Series(['a b a b c', 'a b c'])
    >>> top_n_ngrams(s, top_n=3, ngrams=2)
    (a, b)    3
    (b, c)    2
    (b, a)    1
    dtype: int64
    """
    f = comp(pd.Series, partial(nltk.ngrams, n=ngrams), str.split)
    return s.apply(f).stack().value_counts().head(top_n)


def tokenize(string):
    """
    >>> tokenize("Hey, what's going on?")
    "Hey , what ' s going on ?"
    """
    tokenizer = nltk.tokenize.ToktokTokenizer()
    return tokenizer.tokenize(string, return_str=True)


def stem(string):
    ps = nltk.porter.PorterStemmer()
    stems = [ps.stem(word) for word in string.split()]
    return " ".join(stems)


def lemmatize(string):
    wnl = nltk.stem.WordNetLemmatizer()
    lemmas = [wnl.lemmatize(word) for word in string.split()]
    return " ".join(lemmas)


STOPWORDS = stopword_list = nltk.corpus.stopwords.words("english")


def clean(text: str, ascii=False) -> str:
    if ascii:
        text = text.encode("ascii", "ignore").decode("ascii")
    text = text.strip().lower()
    text = re.sub("\s+", " ", text)
    text = re.sub("[^\s\w]", "", text)
    # contractions
    text = re.sub(r"what's", "what is ", text)
    text = re.sub(r"\'s", " ", text)
    text = re.sub(r"\'ve", " have ", text)
    text = re.sub(r"can't", "can not ", text)
    text = re.sub(r"n't", " not ", text)
    text = re.sub(r"i'm", "i am ", text)
    text = re.sub(r"\'re", " are ", text)
    text = re.sub(r"\'d", " would ", text)
    text = re.sub(r"\'ll", " will ", text)
    text = re.sub(r"\'scuse", " excuse ", text)
    return text



import graphviz
import pandas as pd
from pydataset import data
from sklearn.tree import DecisionTreeClassifier, export_graphviz

import zgulde.ds_util.plotting as zplot
from zgulde.ds_util.modeling import *

# mode aggregation
# tips.groupby('size').agg(stats.mode).apply(lambda s: s.apply(lambda v: v[0][0]))

# better interaction w/ sklearn.metrics.confusion_matrix, e.g. auto labelling
def better_confusion_matrix(actual, predicted, labels=None):
    cm = pd.crosstab(actual, predicted)
    if labels is not None:
        cm.index = pd.Index(labels, name="Actual")
        cm.columns = pd.Index(labels, name="Predictions")
    return cm


def labelled_confusion_matrix():
    matrix = [
        ["TP", "FN (Type II Error)", "Recall = TP / (TP + FN)"],
        ["FP (Type I Error)", "TN", ""],
        ["Precision = TP / (TP + FP)", "", ""],
    ]
    return pd.DataFrame(
        matrix,
        columns=["Predicted +", "Predicted -", ""],
        index=["Actual +", "Actual -", ""],
    )


def rdatasets():
    datasets = data()
    shapes = (
        datasets.apply(lambda row: data(row.dataset_id).shape, axis=1)
        .apply(pd.Series)
        .rename({0: "nrows", 1: "ncols"}, axis=1)
    )
    return pd.concat([datasets, shapes], axis=1)


# https://stackoverflow.com/questions/50559078/generating-random-dates-within-a-given-range-in-pandas
def rand_dates(start, end, n):
    start_u = pd.to_datetime(start).value // 10 ** 9
    end_u = pd.to_datetime(end).value // 10 ** 9
    return pd.DatetimeIndex(
        (10 ** 9 * np.random.randint(start_u, end_u, n)).view("M8[ns]")
    )


def viz_dtree(X, y, **kwargs):
    tree = DecisionTreeClassifier(**kwargs).fit(X, y)
    feature_names = X.columns
    class_names = sorted(y.unique())
    dot = export_graphviz(
        tree,
        out_file=None,
        feature_names=feature_names,
        class_names=class_names,  # target value names
        special_characters=True,
        filled=True,  # fill nodes w/ informative colors
        impurity=False,  # show impurity at each node
        leaves_parallel=True,  # all leaves at the bottom
        proportion=True,  # show percentages instead of numbers at each leaf
        rotate=True,  # left to right instead of top-bottom
        rounded=True,  # rounded boxes and sans-serif font
    )
    graph = graphviz.Source(dot, filename="dtree", format="png")
    graph.view(cleanup=True)



"""
A module that provides all the common data science imports.

For example:

>>> from zgulde.ds_imports import *

and all the common data science imports will be in your namespace.

Intended for quick iteration / experimentation; explicit imports should be used
in production code.
"""

import itertools as it
import re
import sys
from functools import partial, reduce

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import requests as r
import seaborn as sns
import sklearn
from matplotlib import cm
from mpl_toolkits.mplot3d import Axes3D
from pandas import DataFrame, Series
from pydataset import data
from scipy import stats
from scipy.stats import binom, geom, norm, poisson
from sklearn.cluster import DBSCAN, KMeans, dbscan, k_means
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.linear_model import LinearRegression, LogisticRegression, Ridge
from sklearn.metrics import (
    accuracy_score,
    mean_absolute_error,
    mean_squared_error,
    precision_score,
    r2_score,
    recall_score,
)
from sklearn.model_selection import GridSearchCV, cross_val_score, train_test_split
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler, scale
from sklearn.svm import SVC, SVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

import zgulde.extend_pandas
from zgulde.ds_util import *
from zgulde.util import *

# check if we're running interactively
if hasattr(sys, "ps1"):
    # turn on interactive mode in matplotlib
    plt.ion()

tips = data("tips")
mpg = data("mpg")
mtcars = data("mtcars")
swiss = data("swiss")
iris = data("iris").cleanup_column_names()

# plotting style defaults
plt.rc("patch", edgecolor="black", force_edgecolor=True, facecolor="firebrick")
plt.rc("axes", grid=True, fc="#FEFEFE")
plt.rc("grid", linestyle=":", linewidth=0.8, alpha=0.7)
plt.rc("axes.spines", right=False, top=False)
plt.rc("figure", figsize=(11, 8), fc="#FEFEFE")
plt.rc("font", size=12.0)
plt.rc("hist", bins=25)



import pytest

from zgulde.util import *


def test_plucking():
    d = {"x": 1, "y": 2}

    assert pluck(d, "x") == [1]
    assert pluck(d, "y", "x") == [2, 1]

    x, y = pluck(d, "x", "y")
    assert x == 1
    assert y == 2

    assert pluck(d) == []


def test_compose_functions():
    def increment(n):
        return n + 1

    def double(n):
        return n * 2

    assert comp(increment, double)(4) == 9
    assert comp(double, increment)(4) == 10

    assert comp(increment, increment, increment)(4) == 7
    assert comp(increment, double, increment)(4) == 11
    assert comp(increment, increment, double)(4) == 10
    assert comp(double, increment, increment)(4) == 12

    assert comp(double, double, double, double)(4) == 64

    assert comp(increment)(1) == 2

    assert comp()(4) == 4


def test_partition():
    l = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    g = partition(l, 3)
    assert next(g) == [1, 2, 3]
    assert next(g) == [4, 5, 6]
    assert next(g) == [7, 8, 9]
    with pytest.raises(StopIteration):
        next(g)

    assert list(partition(l, 1)) == [[1], [2], [3], [4], [5], [6], [7], [8], [9]]
    assert list(partition(l, 2)) == [[1, 2], [3, 4], [5, 6], [7, 8], [9]]
    assert list(partition(l, 3)) == [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    assert list(partition(l, 4)) == [[1, 2, 3, 4], [5, 6, 7, 8], [9]]

    assert list(partition(l, 8)) == [[1, 2, 3, 4, 5, 6, 7, 8], [9]]
    assert list(partition(l, 9)) == [[1, 2, 3, 4, 5, 6, 7, 8, 9]]
    assert list(partition(l, 10)) == [[1, 2, 3, 4, 5, 6, 7, 8, 9]]

    assert chunk == partition, "It has another name"


def test_pipe():
    def increment(n):
        return n + 1

    def double(n):
        return n * 2

    assert pipe(3) == 3, "a value without any functions will simply be itself"

    assert pipe(3, increment) == 4
    assert pipe(3, double) == 6

    assert pipe(3, increment, increment) == 5
    assert pipe(3, double, double) == 12
    assert pipe(3, increment, double) == 8
    assert pipe(3, double, increment) == 7

    assert pipe(3, increment, double, increment, double) == 18
    assert pipe(3, double, double, increment, increment) == 14
    assert pipe(3, increment, increment, double, double) == 20


def test_and_prev():
    l = [1, 2, 3, 4, 5]

    g = and_prev(l)
    assert next(g) == (None, 1)
    assert next(g) == (1, 2)
    assert next(g) == (2, 3)
    assert next(g) == (3, 4)
    assert next(g) == (4, 5)

    assert list(and_prev(l)) == [(None, 1), (1, 2), (2, 3), (3, 4), (4, 5)]


def test_and_next():
    l = [1, 2, 3, 4, 5]

    g = and_next(l)
    assert next(g) == (1, 2)
    assert next(g) == (2, 3)
    assert next(g) == (3, 4)
    assert next(g) == (4, 5)
    assert next(g) == (5, None)

    assert list(and_next(l)) == [(1, 2), (2, 3), (3, 4), (4, 5), (5, None)]


def test_prev_and_next():
    l = [1, 2, 3]

    g = prev_and_next(l)
    assert next(g) == (None, 1, 2)
    assert next(g) == (1, 2, 3)
    assert next(g) == (2, 3, None)

    assert list(prev_and_next([1, 2, 3])) == [(None, 1, 2), (1, 2, 3), (2, 3, None)]


def test_take():
    l = [1, 2, 3]

    assert take(l, 2) == [1, 2], "takes the first n items"
    assert take(l, 1) == [1], "always returns a list"
    assert take(l, 4) == [1, 2, 3], "giving an n higher than the count is fine"

    def counter():
        n = 0
        while True:
            yield n
            n += 1

    assert take(counter(), 3) == [0, 1, 2], "works with any iterable"
    assert take(it.cycle([1]), 3) == [1, 1, 1], "works with any iterable"


def test_tail():
    assert tail([1, 2, 3], 2) == [2, 3], "takes the last n elements"
    assert tail([1, 2, 3], 1) == [3], "always returns a list"
    assert tail([1, 2, 3], 4) == [1, 2, 3], "giving an n higher than the count is fine"

    assert tail(iter([1, 2, 3, 4]), 2) == [3, 4], "works with iterables"


def test_prepend():
    assert list(prepend([1, 2, 3], 4)) == [
        4,
        1,
        2,
        3,
    ], "adds a single element to the beginning"
    assert list(prepend(it.repeat(1, 2), 0)) == [0, 1, 1], "works with iterables"


def test_drop():
    assert list(drop([1, 2, 3], 1)) == [2, 3], "drops the first n elements"

    xs = it.repeat(0, 5)
    assert list(drop(xs, 3)) == [0, 0], "works with iterables"


def test_append():
    assert list(append([1, 2, 3], 4)) == [
        1,
        2,
        3,
        4,
    ], "appends the given element to the list"
    assert list(append(it.repeat(0, 2), 1)) == [0, 0, 1], "works with iterables"


@pytest.fixture()
def r():
    return MyRange()


def test_it_generates_a_range(r):
    assert r[5:8] == [5, 6, 7, 8]
    assert r[5:8:2] == [5, 7]
    assert r[:5] == [0, 1, 2, 3, 4, 5]
    assert r[5:] == []



message = """
Runnable modules contained here

  zgulde.nbslides     -- (WIP): turn jupyter notebooks into remark.js slides
  zgulde.gitlog       -- make git logs more machine readable
  zgulde.em           -- extract markdown documents from code comments
  zgulde.contact_info
""".strip()

print(message)



import textwrap
from subprocess import check_output

import pytest

from zgulde.em.__main__ import make_categorizor, prev_and_next


@pytest.fixture
def extract_markdown():
    def _extract_markdown(input, expected, args=None):
        cmd = "python -m zgulde.em"
        input = textwrap.dedent(input).strip()
        expected = textwrap.dedent(expected).strip()
        if args is not None:
            cmd += " " + args
        output = check_output(cmd.split(), input=input.encode()).decode().strip()
        print("----- actual")
        print(output)
        print("----- expected")
        print(expected)
        return output, expected

    yield _extract_markdown


def test_categorizor():
    lines = (
        textwrap.dedent(
            """
    #: # Hello, World
    #:
    #: This is a demo.

    foo

    bar

    baz
    """
        )
        .strip()
        .split("\n")
    )
    categorize = make_categorizor("#:")
    categories = [categorize(line) for line in prev_and_next(lines)]
    assert categories == ["docs"] * 3 + ["ignore"] + ["code"] * 5


def test_docs_and_code(extract_markdown):
    actual, expected = extract_markdown(
        """
        #: docs
        #: docs
        #: docs
        code
        code
        """,
        """
        docs
        docs
        docs

        ```
        code
        code
        ```
        """,
    )
    assert actual == expected


def test_just_docs(extract_markdown):
    actual, expected = extract_markdown(
        """
        #: docs
        """,
        """
        docs
        """,
    )
    assert actual == expected


def test_just_code(extract_markdown):
    actual, expected = extract_markdown(
        """
        code
        """,
        """
        ```
        code
        ```
        """,
    )
    assert actual == expected


def test_blank_lines_are_preserved_within_docs_or_code(extract_markdown):
    actual, expected = extract_markdown(
        """
        #: docs
        #:
        #: docs

        code

        code
        """,
        """
        docs

        docs

        ```
        code

        code
        ```
        """,
    )
    assert actual == expected


def test_blank_lines_between_docs_and_code_are_auto_inserted_if_not_present(
    extract_markdown
):
    actual, expected = extract_markdown(
        """
        #: docs
        code
        #: docs
        code
        """,
        """
        docs

        ```
        code
        ```

        docs

        ```
        code
        ```
        """,
    )
    assert actual == expected


def test_multiple_blank_lines_between_docs_and_code_are_ignored(extract_markdown):
    actual, expected = extract_markdown(
        """
        #: docs


        code


        #: docs


        code
        """,
        """
        docs

        ```
        code
        ```

        docs

        ```
        code
        ```
        """,
    )
    assert actual == expected


def test_2_or_more_blanks_in_code_create_seperate_blocks(extract_markdown):
    actual, expected = extract_markdown(
        """
        code


        code
        """,
        """
        ```
        code
        ```

        ```
        code
        ```
        """,
    )
    assert actual == expected


def test_it_highlights_adds_the_given_language_to_fenced_code_blocks(extract_markdown):
    actual, expected = extract_markdown(
        """
        code
        """,
        """
        ```bash
        code
        ```
        """,
        args="-l bash",
    )
    assert actual == expected
    actual, expected = extract_markdown(
        """
        code
        """,
        """
        ```bash
        code
        ```
        """,
        args="--language bash",
    )
    assert actual == expected


def test_it_accepts_a_different_doc_marker(extract_markdown):
    actual, expected = extract_markdown(
        """
        ## Here is some documentation
        this is code
        """,
        """
        Here is some documentation

        ```
        this is code
        ```
        """,
        args="-m ##",
    )
    assert actual == expected
    actual, expected = extract_markdown(
        """
        ## Here is some documentation
        this is code
        """,
        """
        Here is some documentation

        ```
        this is code
        ```
        """,
        args="--doc-marker ##",
    )
    assert actual == expected



import math
from typing import Callable

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from mpl_toolkits.mplot3d import Axes3D
from pydataset import data

import zgulde.extend_pandas

# TODO: refactor some of these to iterate over .groupby instead of unique values

# plotting style defaults -- plt.style.use(style) -- mpl.style.use(style)
style = {
    "animation.html": "html5",
    "axes.facecolor": "#FEFEFE",
    "axes.grid": True,
    "axes.spines.right": False,
    "axes.spines.top": False,
    "figure.facecolor": "#FEFEFE",
    "figure.figsize": (16, 9),
    "font.size": 13.0,
    "grid.alpha": 0.7,
    "grid.linestyle": ":",
    "grid.linewidth": 0.8,
    "hist.bins": 25,
    "patch.edgecolor": "black",
    "patch.facecolor": "firebrick",
    "patch.force_edgecolor": True,
}


def dual_axis(df: pd.DataFrame, x: str) -> Callable:
    """
    >>> mpg = data("mpg")
    >>> dual_axis(mpg, "displ")("hwy")("cty")
    <Figure size ... with 2 Axes>
    """
    fig, ax1 = plt.subplots()
    ax2 = ax1.twinx()

    def plot_y2(y, *args, **kwargs):
        ax2.plot(df[x], df[y], *args, **kwargs)
        return fig

    def plot_y1(y, *args, **kwargs):
        ax1.plot(df[x], df[y], *args, **kwargs)
        return plot_y2

    return plot_y1


def bar_dual_y(
    df: pd.DataFrame,
    x: str,
    y1: str,
    y2: str,
    aggfunc="mean",
    space=0.1,
    ax=None,
    colors=["lightblue", "orange"],
):
    g = df[[x, y1, y2]].groupby(x).agg(aggfunc)

    fig, ax1 = plt.subplots()
    ax2 = ax1.twinx()

    ax1_xticks = [x - (0.5 - space) for x in range(len(g.index))]

    ax1.bar(
        ax1_xticks, g[y1], width=(0.5 - space), align="edge", color=colors[0], label=y1
    )
    ax2.bar(
        g.index, g[y2], width=(0.5 - space), align="edge", label=y2, color=colors[1]
    )

    ax1.set_ylabel(y1)
    ax2.set_ylabel(y2, rotation=270, labelpad=14)
    ax1.set_xlabel(x)
    ax1.set(title=f"{aggfunc} of {y1} and {y2} by {x}")

    fig.legend()

    return fig, (ax1, ax2)


def plot_3d(df, x, y, z, g):
    fig = plt.figure()
    ax = Axes3D(fig)
    for v in sorted(df[g].unique()):
        subset = df[df[g] == v]
        ax.scatter(subset[x], subset[y], subset[z], label=v)
    ax.set(xlabel=x, ylabel=y, zlabel=z)
    ax.legend(title=g)


def hist_by_group(x: pd.Series, g: pd.Series, *args, **kwargs):
    """
    >>> mpg = data('mpg')
    >>> hist_by_group(mpg.hwy, mpg.cyl)
    (<Figure size ... with 4 Axes>, array([[<matplotlib.axes._subplots.AxesSubplot object at ...>,
            <matplotlib.axes._subplots.AxesSubplot object at ...>],
           [<matplotlib.axes._subplots.AxesSubplot object at ...>,
            <matplotlib.axes._subplots.AxesSubplot object at ...>]],
          dtype=object))
    """
    fig, axs = plt.subplots(2, 2) if g.nunique() > 2 else plt.subplots(1, 2)
    fig.suptitle(f"Distribution of {x.name} by {g.name}")
    g = g.top_n(3)
    g.index = x.index

    for ax, v in zip(axs.ravel(), g.unique()):
        x_g = x[g == v]

        ax.hist(x_g, color="pink", *args, **kwargs)
        ax.set_title(v)

        # mean + ci
        xbar = x_g.mean()
        z = 1.96  # 95% ci
        ci = z * (x_g.std() / math.sqrt(x_g.shape[0]))
        ub, lb = xbar + ci, xbar - ci

        ymin, ymax = ax.get_ylim()
        ax.vlines(xbar, ymin, ymax, ls="--", color="gray")
        ax.vlines([lb, ub], ymin, ymax, ls=":", color="gray")

    return fig, axs


def bar_by_group(x: pd.Series, g: pd.Series, aggfunc="mean", *args, **kwargs):
    """
    >>> mpg = data('mpg')
    >>> bar_by_group(mpg.hwy, mpg['class'])
    (<Figure size ... with 1 Axes>, <matplotlib.axes._subplots.AxesSubplot object at ...>)
    """
    g = g.top_n(3)
    fig, ax = plt.subplots()
    x.groupby(g).agg(aggfunc).plot.bar(ax=ax, color="pink", width=1)
    ax.set(title=f"{aggfunc} of {x.name} by {g.name}")
    if aggfunc == "mean":
        xbar = x.agg(aggfunc)
        z = 2.58  # 99% ci
        ci = z * (x.std() / math.sqrt(x.shape[0]))
        ub, lb = xbar + ci, xbar - ci
        ax.hlines(xbar, -0.5, 3.5, ls="--", color="gray")
        ax.hlines([lb, ub], -0.5, 3.5, ls=":", color="gray")
    return fig, ax


def scatter_by_group(df: pd.DataFrame, x: str, y: str, g: str):
    for name, subset in df.groupby(g):
        plt.scatter(subset[x], subset[y], label=name)
    plt.xlabel(x)
    plt.ylabel(y)
    plt.legend(title=g)


def group_proportions(df: pd.DataFrame, x1: str, x2: str, proportions=False, ax=None):
    """
    Visualize the proportion of each group in x2 for each unique group in x1.

    x1 and x2 should both be categorical variables.
    """
    if ax is None:
        fig, ax = plt.subplots()
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
    (
        df.groupby(x1)[x2]
        .apply(pd.Series.value_counts, normalize=proportions)
        .unstack()
        .plot.bar(stacked=True, width=0.9, ax=ax)
    )
    ax.set(ylabel="proportion" if proportions else "count")
    ax.legend(title=x2)
    ax.set_xticklabels(ax.get_xticklabels(), rotation=0)

    return ax


def crosstab_heatmap(
    x: pd.Series,
    y: pd.Series,
    ax=None,
    values=None,
    aggfunc=None,
    normalize=False,
    cmap="Purples",
    fmt=None,
):
    if ax is None:
        fig, ax = plt.subplots()
    if fmt is None:
        fmt = ".2f" if normalize or aggfunc is not None else "d"

    ctab = pd.crosstab(x, y, values=values, aggfunc=aggfunc)
    sns.heatmap(ctab, annot=True, cmap=cmap, fmt=fmt)
    return ax


def crosstab_scatter(
    x: pd.Series, y: pd.Series, ax=None, scale=1000, values=None, aggfunc=None
):
    """
    Visualize the crosstabulation of x and y with a scatter plot where the size
    of the points corresponds to the value of that cell in the crosstabulation.

    For an aggregation other than counting, values and aggfunc can be passed.
    Internally, x, y, values, and aggfunc are all passed to pd.crosstab.
    """
    if ax is None:
        fig, ax = plt.subplots()

    ctab = pd.crosstab(x, y, values=values, aggfunc=aggfunc)
    ctab = ctab.reset_index().melt(id_vars=x.name)
    # min-max scale -- 0 to `scale`
    ctab.value = ((ctab.value - ctab.value.min()) * scale) / (
        ctab.value.max() - ctab.value.min()
    )

    ctab["x_rank"] = ctab[x.name].rank(method="dense")
    ctab["y_rank"] = ctab[y.name].rank(method="dense")

    ax.scatter(ctab.x_rank, ctab.y_rank, s=ctab.value)
    ax.vlines(ctab.x_rank, *ax.get_ylim(), ls="--", color="grey", lw=0.8)
    ax.hlines(ctab.y_rank, *ax.get_xlim(), ls="--", color="grey", lw=0.8)

    ax.set(xticks=ctab.x_rank.unique(), xticklabels=ctab[x.name].unique())
    ax.set(yticks=ctab.y_rank.unique(), yticklabels=ctab[y.name].unique())
    ax.set(xlabel=x.name, ylabel=y.name)

    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)

    return ax


# WIP below
arrowstyles = [
    "-",
    "->",
    "-[",
    "|-|",
    "-|>",
    "<-",
    "<->",
    "<|-",
    "<|-|>",
    "fancy",
    "simple",
    "wedge",
]


def _annotate(s, xy, xytext, style="->"):
    return plt.annotate(s, xy=xy, xytext=xytext, arrowprops={"arrowstyle": style})


def _arrow(fromxy, toxy, style="->"):
    plt.annotate(
        "",
        xy=(11, 0.07),
        xytext=(30, 0.07),
        xycoords="data",
        textcoords="data",
        arrowprops={"arrowstyle": "|-|"},
    )



import argparse
import json
import sys

from .gitlog import get_commits

parser = argparse.ArgumentParser()
parser.prog = "python -m zgulde.gitlog"
parser.description = "Export git log information in JSON format"
parser.add_argument(
    "--repo-path",
    default=".",
    help="path to the repo to be analyzed (default this directory)",
)
parser.add_argument(
    "-o",
    "--outfile",
    type=argparse.FileType("w"),
    default=sys.stdout,
    help="Where to output the processed json data (default stdout)",
)
parser.add_argument(
    "-i",
    "--indent",
    default=2,
    type=int,
    help="Indentation of the resulting json (default: %(default)s)",
)
args = parser.parse_args()

commits = get_commits(args.repo_path)
json.dump(commits, args.outfile, indent=args.indent)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="zgulde",
    version="0.0.41",
    author="Zach Gulde",
    author_email="zachgulde@gmail.com",
    description="A Small Person Utility Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zgulde/zgulde-python",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "pandas>=1.0.1",
        "matplotlib>=2.2.0",
        "scikit-learn>=0.22.1",
        "GitPython==3.1.0",
    ],
)



import math
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
import sklearn
from sklearn.linear_model import LinearRegression, LogisticRegression, Ridge
from sklearn.metrics import (
    accuracy_score,
    mean_absolute_error,
    mean_squared_error,
    precision_score,
    r2_score,
    recall_score,
)
from sklearn.model_selection import GridSearchCV
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.svm import SVC, SVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor


def model(X, y, m, **kwargs):
    return m(**kwargs).fit(X, y)


def residuals(model, X, y):
    return y - model.predict(X)


def tree(df, rformula, **kwargs):
    X, y = df.rformula(rformula)
    if np.issubdtype(y.dtype, np.number):
        return dtree_regressor(X, y, **kwargs)
    else:
        y = y.astype("category").cat.codes
        return dtree_classifier(X, y, **kwargs)


def lr(df, rformula, **kwargs):
    X, y = df.rformula(rformula)
    if np.issubdtype(y.dtype, np.number):
        return linear_regression(X, y, **kwargs)
    else:
        y = y.astype("category").cat.codes
        return logistic_regression(X, y, **kwargs)


def logistic_regression(X, y, **kwargs):
    lr = LogisticRegression(**kwargs).fit(X, y)
    yhat = lr.predict(X)
    coefs = pd.Series(dict(zip(X.columns, lr.coef_[0])))
    return {
        "model": lr,
        "accuracy": accuracy_score(y, yhat),
        "precision": precision_score(y, yhat, average=None),
        "recall": recall_score(y, yhat, average=None),
        "coef": coefs.sort_values(),
    }


def dtree_classifier(X, y, **kwargs):
    tree = DecisionTreeClassifier(**kwargs).fit(X, y)
    yhat = tree.predict(X)
    features = pd.Series(dict(zip(X.columns, tree.feature_importances_)))
    return {
        "model": tree,
        "accuracy": accuracy_score(y, yhat),
        "precision": precision_score(y, yhat, average=None),
        "recall": recall_score(y, yhat, average=None),
        "feature_importances": features.sort_values(),
    }


def dtree_regressor(X, y, **kwargs):
    tree = DecisionTreeRegressor(**kwargs).fit(X, y)
    yhat = tree.predict(X)
    mse = mean_squared_error(y, yhat)
    features = pd.Series(dict(zip(X.columns, tree.feature_importances_)))
    return {
        "model": tree,
        "r2": r2_score(y, yhat),
        "mse": mse,
        "rmse": math.sqrt(mse),
        "mae": mean_absolute_error(y, yhat),
        "feature_importances": features.sort_values(),
    }


def linear_regression(X, y, **kwargs):
    lr = LinearRegression(**kwargs).fit(X, y)
    yhat = lr.predict(X)
    mse = mean_squared_error(y, yhat)
    coefs = pd.Series(dict(zip(X.columns, lr.coef_)))
    return {
        "model": lr,
        "r2": r2_score(y, yhat),
        "mse": mse,
        "rmse": math.sqrt(mse),
        "mae": mean_absolute_error(y, yhat),
        "coef": coefs.sort_values(),
    }


def cv_results_to_df(grid: GridSearchCV) -> pd.DataFrame:
    """
    Presents GridSearchCV results as a data frame

    >>> from pydataset import data
    >>> from sklearn.linear_model import LogisticRegression
    >>> iris = data('iris')
    >>> X, y = iris[['Petal.Length', 'Sepal.Length']], iris.Species
    >>> params = {'penalty': ['none', 'l2'], 'fit_intercept': [True, False]}
    >>> algo = LogisticRegression(random_state=123)
    >>> grid = GridSearchCV(algo, params, cv=8, iid=False).fit(X, y)
    >>> cv_results_to_df(grid)
       fit_intercept penalty     score               model
    0           True    none  0.953216  LogisticRegression
    1           True      l2  0.960161  LogisticRegression
    2          False    none  0.926170  LogisticRegression
    3          False      l2  0.933114  LogisticRegression
    """
    results = grid.cv_results_
    params_and_scores = zip(results["params"], results["mean_test_score"])
    df = pd.DataFrame([dict(**p, score=s) for p, s in params_and_scores])
    df["model"] = grid.best_estimator_.__class__.__name__
    return df


GridCandidates = List[
    Tuple[sklearn.base.BaseEstimator, Dict[str, List[Union[str, int]]]]
]


def multi_grid_search(models: GridCandidates, X, y, cv=4, **kwargs) -> pd.DataFrame:
    """
    combine grid searches for multiple models

    kwargs are passed along to GridSearchCV

    >>> from sklearn.linear_model import Ridge, LinearRegression
    >>> from sklearn.ensemble import RandomForestRegressor
    >>> from pydataset import data
    >>> tips = data('tips')
    >>> X, y = tips[['total_bill', 'size']], tips.tip
    >>> models = []
    >>> models += [(Ridge(), {'alpha': [.1, 1, 10]})]
    >>> models += [(RandomForestRegressor(random_state=123), {'max_depth': range(3, 5), 'n_estimators': [3, 7]})]
    >>> models += [(LinearRegression(), {})]
    >>> multi_grid_search(models, X, y)
       alpha     score                  model  max_depth  n_estimators
    0   10.0  0.458305                  Ridge        NaN           NaN
    1    1.0  0.457351                  Ridge        NaN           NaN
    2    0.1  0.457235                  Ridge        NaN           NaN
    3    NaN  0.457221       LinearRegression        NaN           NaN
    4    NaN  0.325329  RandomForestRegressor        3.0           7.0
    5    NaN  0.294933  RandomForestRegressor        4.0           7.0
    6    NaN  0.272333  RandomForestRegressor        3.0           3.0
    7    NaN  0.254083  RandomForestRegressor        4.0           3.0
    """
    return (
        pd.concat(
            [
                cv_results_to_df(
                    GridSearchCV(model, params, cv=cv, n_jobs=-1, **kwargs).fit(X, y)
                )
                for model, params in models
            ],
            sort=False,
        )
        .sort_values(by="score", ascending=False)
        .reset_index(drop=True)
    )


def inspect_coefs(lm, X) -> Union[pd.Series, pd.DataFrame]:
    """
    View a summary of a linear model's coefficients.

    Works for both linear and logistic regression models.

    >>> import pydataset
    >>> iris = pydataset.data('iris')
    >>> X, y = iris[['Sepal.Length', 'Sepal.Width']], iris.Species
    >>> lm = LogisticRegression(multi_class='auto', solver='lbfgs', random_state=123)
    >>> lm = lm.fit(X, y)
    >>> inspect_coefs(lm, X)
                Sepal.Length  Sepal.Width
    setosa         -2.708902     2.324024
    versicolor      0.612733    -1.570588
    virginica       2.096170    -0.753436

    >>> tips = pydataset.data('tips')
    >>> X, y = tips[['size', 'total_bill']], tips['tip']
    >>> lm = LinearRegression().fit(X, y)
    >>> inspect_coefs(lm, X)
    total_bill    0.092713
    size          0.192598
    dtype: float64
    """
    coef = lm.coef_
    if len(coef.shape) > 1:  # logistic
        return pd.DataFrame(coef, index=lm.classes_, columns=X.columns)
    else:  # linear
        return pd.Series(dict(zip(X.columns, lm.coef_.ravel()))).sort_values()


def inspect_feature_importances(tree, X) -> pd.Series:
    """
    View a summary of a decision tree's feature importances.

    >>> import pydataset
    >>> iris = pydataset.data('iris')
    >>> X, y = iris.drop(columns='Species'), iris.Species
    >>> tree = DecisionTreeClassifier(random_state=123)
    >>> tree = tree.fit(X, y)
    >>> inspect_feature_importances(tree, X)
    Sepal.Width     0.000000
    Sepal.Length    0.026667
    Petal.Width     0.422611
    Petal.Length    0.550723
    dtype: float64
    """
    return pd.Series(dict(zip(X.columns, tree.feature_importances_))).sort_values()


classification_models = [
    (DecisionTreeClassifier(), {"max_depth": range(1, 11)}),
    (KNeighborsClassifier(), {"n_neighbors": range(1, 11)}),
    (
        LogisticRegression(solver="lbfgs", multi_class="auto"),
        {"C": [0.01, 0.1, 1, 10, 100, 1000]},
    ),
]

regression_models = [
    (DecisionTreeRegressor(), {"max_depth": range(1, 11)}),
    (KNeighborsRegressor(), {"n_neighbors": range(1, 11)}),
    (LinearRegression(), {}),
    (Ridge(), {"alpha": [0.01, 0.1, 1, 10, 100, 1000]}),
]



from argparse import ArgumentParser
from os import path, system
from sys import exit
from typing import List, NamedTuple

from IPython import embed


class FlashCard(NamedTuple):
    front: str
    back: str


Deck = List[FlashCard]


def die(msg, code=1):
    print(msg)
    exit(1)


def make_cards(content: str) -> Deck:
    card_contents = content.split("\n===\n")
    faces = [card.strip().split("\n---\n") for card in card_contents]
    cards = [FlashCard(front.strip(), back.strip()) for front, back in faces]
    return cards


def review_cards(cards: Deck):
    for card in cards:
        system("clear")
        print(card.front)
        input("Answer? ")
        print("---")
        print()
        print(card.back)
        input("(c)orrect (i)ncorrect? ")


if __name__ == "__main__":
    parser = ArgumentParser(prog="zgulde.flashcards")
    parser.add_argument("flashcard_file")
    args = parser.parse_args()

    print(args.flashcard_file)

    if not path.exists(args.flashcard_file):
        die(f"Error: could not find file {args.flashcard_file}")

    with open(args.flashcard_file) as f:
        content = f.read()

    cards = make_cards(content)
    review_cards(cards)



"""
This module adds functionality to pandas Series and DataFrame objects. The
objects in pandas will be modified by simply importing this module.

>>> import zgulde.extend_pandas

The following methods are added to all Series:

- `cut`_ (bin): put data into bins; shortcut to pd.cut
- `get_scaler`_: obtain a function that scales a series
- `ln`_: natural log
- `log2`_: log base 2
- `log`_: log base 10
- `outliers`_: detect outliers
- `top_n`_: encode as the most frequent n values or "Other"
- `zscore`_: obtain the z-score for every value

and the following are added to all DataFrames:

- `chi2`_: run chi square tests on all column combinations
- `correlation_heatmap`_: plot a heatmap of the correlations
- `crosstab`_ (xtab): shortcut to pd.crosstab
- `drop_outliers`_: remove outliers
- `get_scalers`_: obtain a function that scales multiple columns
- `hdtl`_: look at the head and tail of the data frame
- `nnull`_ (nna): summarize the number of missing values
- `n_outliers`_: summarize the number of outliers in each numeric column
- `ttest`_: run multiple 1 sample t-tests for multiple categories
- `ttest_2samp`_: run multiple 2 sample t-tests for multiple categories
- `select`_: select and rename columns in a dataframe
- `sql`_: run SQL queries against a dataframe
- `unnest`_: handle multiple values in a single cell

It also defines the left and right shift operators to be similar to
``pandas.DataFrame.pipe``. For example:

>>> import pandas as pd
>>> import numpy as np
>>> df = pd.DataFrame(dict(x=np.arange(4)))
>>> df
   x
0  0
1  1
2  2
3  3
>>> create_y = lambda df: df.assign(y=df.x + 1)
>>> df >> create_y
   x  y
0  0  1
1  1  2
2  2  3
3  3  4
>>> # This gives the same results as .pipe
>>> ((df >> create_y) == df.pipe(create_y)).all(axis=None)
True
"""

import itertools as it
import operator as op
import re
import sqlite3
from functools import partial, reduce
from tempfile import NamedTemporaryFile
from typing import Callable, Generator, List, Tuple, Union, cast

import numpy as np
import pandas as pd
import patsy
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.pyplot import cm
from pandas import DataFrame, Series
from scipy.stats import chi2_contingency, ttest_1samp, ttest_ind

column_name_re = re.compile(r"[^a-zA-Z_0-9]")


def clean_col_name(col: str) -> str:
    col = col.strip().lower().replace(" ", "_").replace(".", "_").replace("-", "_")
    return re.sub(column_name_re, "", col)


def cleanup_column_names(df: DataFrame, inplace=False) -> DataFrame:
    """
    Returns a data frame with the column names cleaned up. Special characters
    are removed and spaces, dots, and dashes are replaced with underscores.

    Parameters
    ----------

    - inplace : Whether or not to modify the data frame in-place and return
                None.

    Example
    -------

    >>> df = pd.DataFrame({'*Feature& A': [1, 2], ' feature.b  ': [2, 3], 'FEATURE-C': [3, 4]})
    >>> df
       *Feature& A   feature.b    FEATURE-C
    0            1             2          3
    1            2             3          4
    >>> df.cleanup_column_names()
       feature_a  feature_b  feature_c
    0          1          2          3
    1          2          3          4
    >>> df
       *Feature& A   feature.b    FEATURE-C
    0            1             2          3
    1            2             3          4
    >>> df.cleanup_column_names(inplace=True)
    >>> df
       feature_a  feature_b  feature_c
    0          1          2          3
    1          2          3          4
    """
    return df.rename(mapper=clean_col_name, axis="columns", inplace=inplace)


def get_scalers(df: DataFrame, columns, **kwargs) -> Callable:
    """
    Obtain a function that will scale multiple columns on a data frame.

    The returned function accepts a data frame and returns the data frame with
    the specified column(s) scaled.

    This can be useful to make sure you apply the same transformation to both
    training and test data sets.

    See the docstring for Series.get_scaler for more details.

    Parameters
    ----------

    - columns : Either a single string, or a list of strings where each string
                is a column name to scale
    - kwargs : any additional arguments are passed to Series.get_scaler for each
               column specified

    Example
    -------

    >>> df = pd.DataFrame(dict(x=[1, 2, 3, 10], y=[-10, 1, 1, 2]))
    >>> df
        x   y
    0   1 -10
    1   2   1
    2   3   1
    3  10   2
    >>> scale_x = df.get_scalers('x', how='minmax')
    >>> scale_x(df)
              x   y
    0  0.000000 -10
    1  0.111111   1
    2  0.222222   1
    3  1.000000   2
    >>> scale_x_and_y = df.get_scalers(['x', 'y'])
    >>> scale_x_and_y(df)
              x         y
    0 -0.734847 -1.494836
    1 -0.489898  0.439658
    2 -0.244949  0.439658
    3  1.469694  0.615521
    >>> df.pipe(scale_x_and_y)
              x         y
    0 -0.734847 -1.494836
    1 -0.489898  0.439658
    2 -0.244949  0.439658
    3  1.469694  0.615521
    """
    if type(columns) is str:  # allow either a single string or a list of strings
        columns = [columns]
    scalers = [df[col].get_scaler(**kwargs) for col in columns]
    return partial(reduce, lambda df, f: df.pipe(f), scalers)


def qcut(s: Series, *args, **kwargs):
    """
    Bin series values into discrete intervals.

    Shortcut to pd.cut

    Parameters
    ----------

    - args : positional arguments passed to ``pandas.cut``
    - keyword : keyword arguments passed to ``pandas.cut``

    Example
    -------

    >>> x = pd.Series(range(1, 7))
    >>> x
    0    1
    1    2
    2    3
    3    4
    4    5
    5    6
    dtype: int64
    >>> x.qcut(2)
    0    (0.999, 3.5]
    1    (0.999, 3.5]
    2    (0.999, 3.5]
    3      (3.5, 6.0]
    4      (3.5, 6.0]
    5      (3.5, 6.0]
    dtype: category
    Categories (2, interval[float64]): [(0.999, 3.5] < (3.5, 6.0]]
    """
    return pd.qcut(s, *args, **kwargs)


def cut(s: Series, *args, **kwargs):
    """
    Bin series values into discrete intervals.

    Shortcut to pd.cut

    Parameters
    ----------

    - args : positional arguments passed to ``pandas.cut``
    - keyword : keyword arguments passed to ``pandas.cut``

    Example
    -------

    >>> x = pd.Series(range(1, 7))
    >>> x
    0    1
    1    2
    2    3
    3    4
    4    5
    5    6
    dtype: int64
    >>> x.cut(2)
    0    (0.995, 3.5]
    1    (0.995, 3.5]
    2    (0.995, 3.5]
    3      (3.5, 6.0]
    4      (3.5, 6.0]
    5      (3.5, 6.0]
    dtype: category
    Categories (2, interval[float64]): [(0.995, 3.5] < (3.5, 6.0]]
    >>> x.cut(bins=[0, 3, 6])
    0    (0, 3]
    1    (0, 3]
    2    (0, 3]
    3    (3, 6]
    4    (3, 6]
    5    (3, 6]
    dtype: category
    Categories (2, interval[int64]): [(0, 3] < (3, 6]]
    """
    return pd.cut(s, *args, **kwargs)


def get_scaler(s: Series, how="zscore"):
    """
    Obtain a function that will scale the series on a data frame.

    The returned function accepts a data frame and returns the data frame with
    the specified column scaled.

    This can be useful to make sure you apply the same transformation to both
    training and test data sets.

    - zscore = (x - mu) / sigma
    - minmax = (x - min) / (max - min)

    Parameters
    ----------

    - how : One of {'zscore', 'minmax'} to either apply z-score or min-max
            scaling

    Example
    -------

    >>> df = pd.DataFrame(dict(x=[1, 2, 3, 4, 5, 1000], y=[1000, 2, 3, 4, 5, 6]))
    >>> scale_x = df.x.get_scaler()
    >>> scale_x(df)
              x     y
    0 -0.413160  1000
    1 -0.410703     2
    2 -0.408246     3
    3 -0.405789     4
    4 -0.403332     5
    5  2.041229     6
    >>> scale_y = df.y.get_scaler('minmax')
    >>> scale_y(df)
          x         y
    0     1  1.000000
    1     2  0.000000
    2     3  0.001002
    3     4  0.002004
    4     5  0.003006
    5  1000  0.004008
    >>> df.pipe(scale_x).pipe(scale_y)
              x         y
    0 -0.413160  1.000000
    1 -0.410703  0.000000
    2 -0.408246  0.001002
    3 -0.405789  0.002004
    4 -0.403332  0.003006
    5  2.041229  0.004008
    """
    name = s.name
    if how == "zscore":
        mu = s.mean()
        sigma = s.std()

        def scale(df: DataFrame) -> DataFrame:
            scaled_series = (df[name] - mu) / sigma
            kwargs = {name: scaled_series}
            return df.assign(**kwargs)

        return scale
    elif how == "minmax":
        min = s.min()
        max = s.max()

        def scale(df: DataFrame) -> DataFrame:
            scaled_series = (df[name] - min) / (max - min)
            kwargs = {name: scaled_series}
            return df.assign(**kwargs)

        return scale
    raise ValueError("how must be one of {zscore, minmax}")


def zscore(s: Series) -> Series:
    """
    Returns the z-score for every value in the series.

    Z = (x - mu) / sigma

    Example
    -------

    >>> x = pd.Series([1, 2, 3, 4, 5, 6, 7, 8, 9])
    >>> x
    0    1
    1    2
    2    3
    3    4
    4    5
    5    6
    6    7
    7    8
    8    9
    dtype: int64
    >>> x.zscore()
    0   -1.460593
    1   -1.095445
    2   -0.730297
    3   -0.365148
    4    0.000000
    5    0.365148
    6    0.730297
    7    1.095445
    8    1.460593
    dtype: float64
    """
    return (s - s.mean()) / s.std()


def drop_outliers(df: DataFrame, cols: Union[str, List[str]], **kwargs) -> Series:
    """
    Drop rows with outliers in the given columns from the dataframe

    See the docs for .outliers for more details on parameters, and to customize
    how the outliers are detected.

    Parameters
    ----------

    cols : either a string or a list of strings of which column(s) to drop the
           outliers in
    kwargs : additional key-word arguments passed on to
             ``pandas.Series.outliers``

    Example
    -------

    >>> df = pd.DataFrame(dict(x=[1, 2, 3, 4, 5, 1000], y=[1000, 2, 3, 4, 5, 6]))
    >>> df
          x     y
    0     1  1000
    1     2     2
    2     3     3
    3     4     4
    4     5     5
    5  1000     6
    >>> df.drop_outliers('x')
       x     y
    0  1  1000
    1  2     2
    2  3     3
    3  4     4
    4  5     5
    >>> df.drop_outliers('y')
          x  y
    1     2  2
    2     3  3
    3     4  4
    4     5  5
    5  1000  6
    >>> df.drop_outliers(['x', 'y'])
       x  y
    1  2  2
    2  3  3
    3  4  4
    4  5  5
    """
    if type(cols) is str:
        cols = [cast(str, cols)]
    to_keep = [~df[col].outliers(**kwargs) for col in cols]
    return df[list(reduce(op.and_, to_keep))]


def n_outliers(df: DataFrame, **kwargs) -> Series:
    """
    Provide a summary of the number of outliers in each numeric column.

    Parameters
    ----------

    - kwargs : any additional arguments to pass along to
               ``pandas.Series.outliers``

    Returns
    -------

    A ``pandas.DataFrame`` indexed by the column names of the the data frame,
    with columns that indicate the number of outliers and the percentage of
    outliers in each column.

    Example
    -------

    >>> x = [1, 2, 3, 4, 5, 1, 2, 3, 4, 5]
    >>> y = [1, 2, 3, 4, 5, 100, 2, 3, 4, 5]
    >>> z = [1, 2, 3, 4, 5, -100, 2, 3, 100, 5]
    >>> df = pd.DataFrame(dict(x=x, y=y, z=z))
    >>> df
       x    y    z
    0  1    1    1
    1  2    2    2
    2  3    3    3
    3  4    4    4
    4  5    5    5
    5  1  100 -100
    6  2    2    2
    7  3    3    3
    8  4    4  100
    9  5    5    5
    >>> df.n_outliers()
       n_outliers  p_outliers
    x           0         0.0
    y           1         0.1
    z           2         0.2
    """
    n_outliers = df.select_dtypes("number").apply(lambda s: s.outliers(**kwargs).sum())
    p_outliers = n_outliers / df.shape[0]
    return pd.DataFrame(dict(n_outliers=n_outliers, p_outliers=p_outliers))


def outliers(s: Series, how="iqr", k=1.5, std_cutoff=2) -> Series:
    """
    Detect outliers in the series.

    Returns
    -------

    A pandas Series of boolean values indicating whether each point is an
    outlier or not.

    Parameters
    ----------

    how : {'iqr', 'std'}, default 'iqr'
        - 'iqr' : identify outliers based on whether they are > q3 + k * iqr
          or < q1 - k * iqr
        - 'std' : identify outliers based on whether they are further than a
                  specified number of standard deviations from the mean
    k : value to multiply the iqr by for outlier detection. Ignored when
        how='std'. Default 1.5
    std_cutoff : cutoff for identifying an outlier based on standard deviation.
                 Ignored when how='iqr'. Default 2

    Example
    -------

    >>> df = pd.DataFrame(dict(x=[1, 2, 3, 4, 5, 6, 100],
    ...                        y=[-100, 5, 3, 4, 1, 2, 0]))
    >>> df
         x    y
    0    1 -100
    1    2    5
    2    3    3
    3    4    4
    4    5    1
    5    6    2
    6  100    0
    >>> df.x.outliers()
    0    False
    1    False
    2    False
    3    False
    4    False
    5    False
    6     True
    Name: x, dtype: bool
    >>> df[df.x.outliers()]
         x  y
    6  100  0
    """
    if how == "iqr":
        q1 = s.quantile(0.25)
        q3 = s.quantile(0.75)
        iqr = q3 - q1
        return (s < q1 - k * iqr) | (s > q3 + k * iqr)
    elif how == "std":
        return zscore(s).abs() > std_cutoff
    raise ValueError("how must be one of {iqr,std}")


def nnull(df: DataFrame, axis=0) -> DataFrame:
    """
    Provide a summary of null values in each column.

    alias of nna

    Example
    -------

    >>> df = pd.DataFrame(dict(x=[1, 2, np.nan], y=[4, np.nan, np.nan]))
    >>> df
         x    y
    0  1.0  4.0
    1  2.0  NaN
    2  NaN  NaN
    >>> nulls_by_column = df.nnull()
    >>> nulls_by_column
       n_missing  p_missing
    x          1   0.333333
    y          2   0.666667
    >>> nulls_by_row = df.nnull(axis=1)
    >>> nulls_by_row
       n_missing  p_missing
    0          0        0.0
    1          1        0.5
    2          2        1.0
    """
    n_missing = df.isnull().sum(axis=axis)
    p_missing = n_missing / df.shape[axis]
    return pd.DataFrame(dict(n_missing=n_missing, p_missing=p_missing))


def unnest(df: DataFrame, col: str, split=True, sep=",", reset_index=True) -> DataFrame:
    """
    Turns a column with multiple values in each row in it into separate rows,
    each with a single value.

    Parameters
    ----------

    - col : name of the column to unnest
    - split : default True. whether or not to split the data in the column. Set
              this to False if the column already contains lists in each row
    - sep : separator to split on. Ignored if split=False
    - reset_index : default True. whether to reset the index in the resulting
                    data frame. If False, the resulting data frame will have an
                    index that could contain duplicates.

    Example
    -------

    >>> df = pd.DataFrame(dict(x=list('abc'), y=['a,b,c', 'd,e', 'f']))
    >>> df
       x      y
    0  a  a,b,c
    1  b    d,e
    2  c      f
    >>> df.unnest('y')
       x  y
    0  a  a
    1  a  b
    2  a  c
    3  b  d
    4  b  e
    5  c  f
    """
    s = df[col].str.split(sep) if split else df[col]
    s = s.apply(pd.Series).stack().reset_index(level=1, drop=True)

    s.name = col

    return (
        df.drop(columns=[col])
        .join(s)
        .pipe(lambda df: df.reset_index(drop=True) if reset_index else df)
    )


def correlation_heatmap(df: DataFrame, fancy=False, **kwargs):
    """
    Plot a heatmap of the correlation matrix for the data frame.

    Any additional kwargs are passed to ``seaborn.heatmap`` and the resulting
    axes object is returned.

    >>> x = np.arange(0, 10)
    >>> y = x / 2
    >>> df = pd.DataFrame(dict(x=x, y=y))
    >>> df.correlation_heatmap()
    <matplotlib.axes._subplots.AxesSubplot object at ...>
    """
    if "cmap" not in kwargs:
        kwargs["cmap"] = cm.coolwarm_r
        # kwargs['cmap'] = cm.PiYG

    if not fancy:
        return sns.heatmap(df.corr(), center=0, annot=True, **kwargs)

    cmat = df.corr()

    sns.set(style="white")
    mask = np.zeros_like(cmat, dtype=np.bool)
    mask[np.triu_indices_from(mask)] = True

    # Draw the heatmap with the mask and correct aspect ratio
    plt.figure(figsize=(12, 12))
    sns.heatmap(
        cmat,
        mask=mask,
        cbar=True,
        annot=True,
        square=True,
        fmt=".2f",
        annot_kws={"size": 10},
        **kwargs
    )
    plt.yticks(rotation=0)
    return plt


def hdtl(df: DataFrame, n=3) -> DataFrame:
    """
    Return the head and the tail of the data frame.

    Parameters
    ----------

    - n : number of rows to get from both the head and tail

    Example
    -------

    >>> df = pd.DataFrame(dict(x=np.arange(10), y=np.arange(10)))
    >>> df
       x  y
    0  0  0
    1  1  1
    2  2  2
    3  3  3
    4  4  4
    5  5  5
    6  6  6
    7  7  7
    8  8  8
    9  9  9
    >>> df.hdtl(1)
       x  y
    0  0  0
    9  9  9
    >>> df.hdtl()
       x  y
    0  0  0
    1  1  1
    2  2  2
    7  7  7
    8  8  8
    9  9  9
    """
    return pd.concat([df.head(n), df.tail(n)])


def log(s: Series):
    """
    Returns the log base 10 of the values in the series using np.log10

    Example
    -------

    >>> x = pd.Series([1, 10, 100, 1000])
    >>> x
    0       1
    1      10
    2     100
    3    1000
    dtype: int64
    >>> x.log()
    0    0.0
    1    1.0
    2    2.0
    3    3.0
    dtype: float64
    """
    return np.log10(s)


def ln(s: Series):
    """
    Returns the natural log of the values in the series using np.log

    Example
    -------

    >>> x = pd.Series([1, np.e, np.e ** 2, np.e ** 3])
    >>> x
    0     1.000000
    1     2.718282
    2     7.389056
    3    20.085537
    dtype: float64
    >>> x.ln()
    0    0.0
    1    1.0
    2    2.0
    3    3.0
    dtype: float64
    """
    return np.log(s)


def log2(s: Series):
    """
    Returns the log base 2 of the values in the series using np.log2

    Example
    -------

    >>> x = pd.Series([1, 2,4, 8, 16])
    >>> x
    0     1
    1     2
    2     4
    3     8
    4    16
    dtype: int64
    >>> x.log2()
    0    0.0
    1    1.0
    2    2.0
    3    3.0
    4    4.0
    dtype: float64
    """
    return np.log2(s)


def crosstab(df: DataFrame, rows, cols, values=None, **kwargs) -> DataFrame:
    """
    Shortcut to call to pd.crosstab.

    Parameters
    ----------

    - rows : the name of the columns that will make up the rows in resulting
             contingency table
    - cols : the name of the columns that will make up the columns in resulting
             contingency table
    - values : (optional) name of the column to use for the cell values in the
               resulting contingency table. If supplied, aggfunc must be
               provided as well. See ``pd.crosstab`` for more details.
    - kwargs : any additional key word arguments to pass along to
               ``pd.crosstab``

    Example
    -------

    >>> df = pd.DataFrame(dict(x=list('aaabbb'), y=list('cdcdcd'), z=range(6)))
    >>> df
       x  y  z
    0  a  c  0
    1  a  d  1
    2  a  c  2
    3  b  d  3
    4  b  c  4
    5  b  d  5
    >>> df.crosstab('x', 'y')
    y  c  d
    x
    a  2  1
    b  1  2
    >>> (df.crosstab('x', 'y') == pd.crosstab(df.x, df.y)).all(axis=None)
    True
    >>> df.crosstab('x', 'y', margins=True)
    y    c  d  All
    x
    a    2  1    3
    b    1  2    3
    All  3  3    6
    >>> df.xtab(rows='x', cols='y', values='z', aggfunc='mean')
    y  c  d
    x
    a  1  1
    b  4  4
    """
    if values is not None:
        kwargs["values"] = df[values]
    return pd.crosstab(df[rows], df[cols], **kwargs)


def ttest(df: DataFrame, target: str) -> DataFrame:
    """
    Runs a 1 sample t-test comparing the specified target variable to the
    overall mean among all of the possible subgroups.

    Parameters
    ----------

    - target : name of the column that holds the target (continuous) variable

    Example
    -------

    >>> from seaborn import load_dataset
    >>> tips = load_dataset('tips')
    >>> tips = tips[['total_bill', 'day', 'time']]
    >>> tips.ttest('total_bill')
                     statistic    pvalue    n
    variable value
    day      Sun      1.603035  0.113130   76
             Sat      0.644856  0.520737   87
             Thur    -2.099957  0.039876   62
             Fri     -1.383042  0.183569   19
    time     Dinner   1.467432  0.144054  176
             Lunch   -2.797882  0.006710   68
    """
    results = []
    for col in df.drop(columns=target):
        unique_vals = df[col].unique()
        ttests = DataFrame(
            [
                ttest_1samp(df[df[col] == v][target], df[target].mean())
                for v in unique_vals
            ]
        )
        ns = [df[df[col] == v].shape[0] for v in unique_vals]
        ttests = ttests.assign(n=ns, value=unique_vals, variable=col)
        results.append(ttests)
    return pd.concat(results, axis=0).set_index(["variable", "value"])


def ttest_2samp(df: DataFrame, target: str) -> DataFrame:
    """
    Runs a 2 sample t-test comparing the specified target variable for every
    unique value from every other column in the data frame.

    The resulting t-statistic and pvalue are based on subdividing the data for
    each unique value for each column, with each individual value indicating
    that the test was performed based on belonging to that unique value vs not
    belonging to that group.

    Parameters
    ----------

    - target : name of the column that holds the target (continuous) variable

    Example
    -------

    >>> from seaborn import load_dataset
    >>> tips = load_dataset('tips')
    >>> tips = tips[['total_bill', 'day', 'time']]
    >>> tips.ttest_2samp('total_bill')
                     statistic    pvalue    n
    variable value
    day      Sun      1.927317  0.055111   76
             Sat      0.855634  0.393046   87
             Thur    -2.170294  0.030958   62
             Fri     -1.345462  0.179735   19
    time     Dinner   2.897638  0.004105  176
             Lunch   -2.897638  0.004105   68
    """
    results = []
    for col in df.drop(columns=target):
        unique_vals = df[col].unique()
        ttests = DataFrame(
            [
                ttest_ind(df[df[col] == v][target], df[df[col] != v][target])
                for v in unique_vals
            ]
        )
        ns = [df[df[col] == v].shape[0] for v in unique_vals]
        ttests = ttests.assign(n=ns, value=unique_vals, variable=col)
        results.append(ttests)
    return pd.concat(results, axis=0).set_index(["variable", "value"])


def chi2(df: DataFrame) -> Tuple[DataFrame, DataFrame]:
    """
    Performs a chi squared contingency table test between all the combinations
    of two columns in the data frame.

    Returns
    -------

    (pvals, chi2s)

    A tuple with two data frames, each which have all of the columns from the
    original data frame as both the indexes and the columns. The values in the
    first are the p-values, the values in the second are the chi square test
    statistics.

    Example
    -------

    >>> from seaborn import load_dataset
    >>> tips = load_dataset('tips')
    >>> p_vals, chi2s = tips[['smoker', 'time', 'day']].chi2()
    >>> p_vals
                 smoker        time          day
    smoker          NaN    0.477149  1.05676e-05
    time       0.477149         NaN   8.4499e-47
    day     1.05676e-05  8.4499e-47          NaN
    >>> chi2s
              smoker      time      day
    smoker       NaN  0.505373  25.7872
    time    0.505373       NaN  217.113
    day      25.7872   217.113      NaN
    """
    p_vals = pd.DataFrame(index=df.columns, columns=df.columns)
    chi2s = p_vals.copy()
    for x1, x2 in it.combinations(df.columns, 2):
        stat, p, *_ = chi2_contingency(df.xtab(x1, x2))
        p_vals.loc[x1, x2] = p
        p_vals.loc[x2, x1] = p
        chi2s.loc[x1, x2] = stat
        chi2s.loc[x2, x1] = stat
    return p_vals, chi2s


def rformula(df, formula):
    """
    Split a data frame into X and y based on an R Formula.

    Based on patsy formulas. See
    https://patsy.readthedocs.io/en/latest/formulas.html for valid formulas.

    Returns
    -------

    A tuple where the first element is a pandas DataFrame containing the
    independent variables, and the second is a pandas Series containing the
    dependent variable.

    Example
    -------

    >>> df = pd.DataFrame(dict(a=[1, 2, 3], b=[4, 5, 6], c=[7, 8, 9]))
    >>> X, y = df.rformula('a ~ b')
    >>> X
       b
    0  4.0
    1  5.0
    2  6.0
    >>> y
         a
    0  1.0
    1  2.0
    2  3.0
    >>> X, y = df.rformula('c ~ a + b')
    >>> X
       a    b
    0  1.0  4.0
    1  2.0  5.0
    2  3.0  6.0
    >>> y
         c
    0  7.0
    1  8.0
    2  9.0
    >>> X, y = df.rformula('b ~ a + a:c')
    >>> X
         a   a:c
    0  1.0   7.0
    1  2.0  16.0
    2  3.0  27.0
    >>> y
         b
    0  4.0
    1  5.0
    2  6.0
    >>> X, y = df.rformula('b ~ a*c')
    >>> X
         a    c   a:c
    0  1.0  7.0   7.0
    1  2.0  8.0  16.0
    2  3.0  9.0  27.0
    >>> y
         b
    0  4.0
    1  5.0
    2  6.0
    """
    y, X = patsy.dmatrices(formula, df, return_type="dataframe")
    return X.drop(columns="Intercept"), y


def pipe(df: DataFrame, fn: Callable):
    return df.pipe(fn)


def sql(
    df: DataFrame, query: str, table="df", in_memory=True, index=False
) -> pd.DataFrame:
    """
    Run a SQL query against the dataframe.

    The dataframe is converted to a sqlite database and the provided query is
    run against it. As such, any SQL that is valid in sqlite is supported.

    Parameters
    ----------

    - query: The SQL query to run
    - table: (optional) name of the table to call the dataframe. Defaults to `df`
    - in_memory: whether or not the sqlite database should be in memory or
                 written to an external file. If False, a temporary file will be
                 created and cleaned up after the query is run.
    - index: whether or not to include the dataframe's index in the table that
             is queried against.

    Examples
    --------

    >>> df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6], 'c': ['a', 'a', 'b']})
    >>> df
       a  b  c
    0  1  4  a
    1  2  5  a
    2  3  6  b
    >>> df.sql('SELECT * FROM df')
       a  b  c
    0  1  4  a
    1  2  5  a
    2  3  6  b
    >>> df.sql('SELECT * FROM my_df', table='my_df')
       a  b  c
    0  1  4  a
    1  2  5  a
    2  3  6  b
    >>> df.sql('SELECT c, AVG(a + b) FROM df GROUP BY c')
       c  AVG(a + b)
    0  a         6.0
    1  b         9.0
    """
    if in_memory:
        conn = sqlite3.connect(":memory:")
        df.to_sql(table, conn, index=index)
        return pd.read_sql(query, conn)
    else:
        with NamedTemporaryFile() as f:
            conn = sqlite3.connect(f.name)
            df.to_sql(table, conn, index=index)
            return pd.read_sql(query, conn)


def top_n(s: pd.Series, n=3, other_val="Other") -> pd.Series:
    """
    Convert the series to the most frequent n values and use other_val for the
    rest.

    Parameters
    ----------

    - n: number of values to consider
    - other_val: value that will be used for everything that isn't the most
                 frequent n values

    Returns
    -------

    A pandas Series

    Examples
    --------

    >>> s = pd.Series(['a', 'a', 'b', 'b', 'c', 'd'])
    >>> s.top_n(2)
    0        a
    1        a
    2        b
    3        b
    4    Other
    5    Other
    dtype: object
    """
    top_n = s.value_counts().index[:n]
    return pd.Series(np.where(s.isin(top_n), s, other_val), name=s.name)


def pluck(df: pd.DataFrame, *cols: str) -> List[pd.Series]:
    """
    >>> df = pd.DataFrame({'x': [1, 2, 3], 'y': ['a', 'b', 'c']})
    >>> x, y = df.pluck('x', 'y')
    >>> x
    0    1
    1    2
    2    3
    Name: x, dtype: int64
    >>> y
    0    a
    1    b
    2    c
    Name: y, dtype: object
    """
    return [df[col] for col in cols]


def select(df, *args: str, **kwargs: str) -> pd.DataFrame:
    """
    Return specified columns from a dataframe, optionally renaming some.

    Parameters
    ----------

    - args: strings that are column names to include
    - kwargs: any additional columns to rename

    Returns
    -------

    A subset of the dataframe

    Examples
    --------

    >>> df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6], 'c': [7, 8, 9]})
    >>> df
       a  b  c
    0  1  4  7
    1  2  5  8
    2  3  6  9
    >>> df.select('a')
       a
    0  1
    1  2
    2  3
    >>> df.select('a', 'b')
       a  b
    0  1  4
    1  2  5
    2  3  6
    >>> df.select('a', c='the_letter_c')
       a  the_letter_c
    0  1             7
    1  2             8
    2  3             9
    >>> df.select(a='A', b='BBB')
       A  BBB
    0  1    4
    1  2    5
    2  3    6
    """
    return df[[*args, *kwargs.keys()]].rename(columns=kwargs)


pd.Series.bin = cut
pd.Series.cut = cut
pd.Series.get_scaler = get_scaler
pd.Series.ln = ln
pd.Series.log = log
pd.Series.log2 = log2
pd.Series.outliers = outliers
pd.Series.qcut = qcut
pd.Series.top_n = top_n
pd.Series.zscore = zscore

pd.DataFrame.cleanup_column_names = cleanup_column_names
pd.DataFrame.chi2 = chi2
pd.DataFrame.correlation_heatmap = correlation_heatmap
pd.DataFrame.crosstab = crosstab
pd.DataFrame.drop_outliers = drop_outliers
pd.DataFrame.get_scalers = get_scalers
pd.DataFrame.hdtl = hdtl
pd.DataFrame.__lshift__ = pipe
pd.DataFrame.n_outliers = n_outliers
pd.DataFrame.nna = nnull
pd.DataFrame.nnull = nnull
pd.DataFrame.pluck = pluck
pd.DataFrame.__rshift__ = pipe
pd.DataFrame.rformula = rformula
pd.DataFrame.ttest_2samp = ttest_2samp
pd.DataFrame.ttest = ttest
pd.DataFrame.unnest = unnest
pd.DataFrame.xtab = crosstab
pd.DataFrame.select = select
pd.DataFrame.sql = sql

series_extensions = [cut, get_scaler, ln, log, log2, outliers, qcut, top_n, zscore]

data_frame_extensions = [
    chi2,
    cleanup_column_names,
    correlation_heatmap,
    crosstab,
    drop_outliers,
    get_scalers,
    hdtl,
    nnull,
    n_outliers,
    pluck,
    ttest,
    ttest_2samp,
    unnest,
    select,
    sql,
]



import logging
import re
import sys
from os import path
from subprocess import check_output

import pandas as pd

commit_re = re.compile(
    r"""
^(?P<author>.*?)
\s
(?P<timestamp>\d+-\d+-\d+\s\d+:\d+:\d+\s-\d+)
\s
(?P<message>.*)$
""",
    re.VERBOSE,
)


def get_commits(repo_path: str) -> str:
    log_format = "%n###%n%ae %ai %s"
    command = ["git", "-C", repo_path, "log", "--stat", f"--pretty={log_format}"]
    logging.debug("[wrangle.py#get_commits] command=" + " ".join(command))
    return check_output(command).decode("utf8").strip()


def handle_section(text):
    lines = text.split("\n")
    data = re.match(commit_re, lines[0]).groupdict()

    if len(lines) == 1:
        return data

    lines = "\n".join(lines[1:]).strip().split("\n")
    files_changed = lines[-1].strip()

    data["n_files_changed"] = int(re.search(r"^\d+", files_changed[0])[0])

    if "insertions" in files_changed:
        data["insertions"] = int(re.search("(\d+)\sinsertions", files_changed)[1])
    if "deletions" in files_changed:
        data["deletions"] = int(re.search("(\d+)\sdeletions", files_changed)[1])

    return data


def get_commit_df(repo_path, use_cache=True) -> pd.DataFrame:
    contents = get_commits(repo_path)
    sections = [
        section.strip() for section in contents.split("###\n") if section.strip()
    ]
    df = pd.DataFrame([handle_section(section) for section in sections])

    for col in "n_files_changed", "insertions", "deletions":
        df[col] = df[col].fillna(0).astype(int)

    df.timestamp = pd.to_datetime(df.timestamp, utc=True)
    df = df.set_index("timestamp").tz_convert("America/Chicago").sort_index()

    return df


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.prog = "python -m zgulde.gitlog"
    parser.description = "Export git log information as a csv"
    parser.add_argument(
        "--repo-path",
        help="Local path to the git repo to be analyzed (default this directory)",
        default=".",
    )
    parser.add_argument(
        "-v", "--verbose", help="increase output verbosity", action="count"
    )
    args = parser.parse_args()

    if args.verbose is None:
        loglevel = logging.WARN
    elif args.verbose == 1:
        loglevel = logging.INFO
    elif args.verbose >= 2:
        loglevel = logging.DEBUG
    logging.basicConfig(
        format="%(asctime)s:%(levelname)s:%(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=loglevel,
    )

    df = get_commit_df(args.repo_path)
    print(df.to_csv(index=True))
